import pytorch_lightning as pl
import torch


class DCLightningDatasetBatch:

    def __init__(self, batch):
        X = [batch[0]]
        y = [batch[1]]
        w = [batch[2]]
        self.batch_list = [X, y, w]


def collate_dataset_wrapper(batch):
    return DCLightningDatasetBatch(batch)


class DCLightningDatasetModule(pl.LightningDataModule):
    """DeepChem Lightning Dataset Module to be used with the DCLightningModule and a Lightning trainer.

    This module wraps over the the deepchem pytorch dataset and dataloader providing a generic interface to run training.
    https://pytorch-lightning.readthedocs.io/en/stable/extensions/datamodules.html

    Notes
    -----
    This class requires PyTorch to be installed.
    """

    def __init__(self, dataset, batch_size, collate_fn):
        """Create a new DCLightningDatasetModule.

        Parameters
        ----------
        dataset: A deepchem dataset.
        batch_size: Batch size for the dataloader.
        collate_fn: Method to collate instances across batch.
        """
        super().__init__()
        self._batch_size = batch_size
        self._dataset = dataset.make_pytorch_dataset(
            batch_size=self._batch_size)
        self.collate_fn = collate_fn

    def setup(self, stage):
        self.train_dataset = self._dataset

    def train_dataloader(self):
        """Returns the train dataloader from train dataset.

        Returns
        -------
        dataloader: train dataloader to be used with DCLightningModule.
        """
        dataloader = torch.utils.data.DataLoader(
            self.train_dataset,
            batch_size=None,
            collate_fn=self.collate_fn,
            shuffle=False,
        )
        return dataloader



import os
import deepchem as dc


def test_batch_coulomb_matrix_features():
    # Get Data
    current_dir = os.path.dirname(os.path.abspath(__file__))
    dataset_file = os.path.join(current_dir, 'assets/qm9_mini.sdf')
    TASKS = ["alpha", "homo"]
    loader = dc.data.SDFLoader(tasks=TASKS,
                               featurizer=dc.feat.CoulombMatrix(29),
                               sanitize=True)
    data = loader.create_dataset(dataset_file, shard_size=100)
    inputs = dc.utils.batch_utils.batch_coulomb_matrix_features(data.X)

    # Checks that all atoms exits in array
    assert inputs[0].shape == (149,)

    # Checks shape of gaussian distance
    assert inputs[1].shape == (1215, 100)

    # Checks all molecule membership exist
    for i in range(0, 21):
        if i not in inputs[2]:
            raise AssertionError("All molecules not in the matrix")

    # Check Distance Membership shape
    assert inputs[3].shape == (1215,)
    assert inputs[4].shape == (1215,)



"""
Test for electron_sampler.py
"""

import numpy as np
from deepchem.utils.electron_sampler import ElectronSampler


def f(x):
    # dummy function which can be passed as the parameter f to simultaneous_move and single_move
    return 2 * np.log(np.random.uniform(low=0, high=1.0, size=np.shape(x)[0]))


def test_mean():
    distribution = ElectronSampler(np.array([[1, 1, 3], [3, 2, 3]]), f)
    x1 = np.array([[[[1, 2, 3]]], [[[4, 5, 6]]]])
    mean = distribution.harmonic_mean(x1)
    assert (mean == np.array([[[[1.3333333333333333]]],
                              [[[4.988597077109626]]]])).all()


def test_log_prob():
    x1 = np.array([[[[1, 2, 3]]], [[[4, 5, 6]]]])
    x2 = np.array([[[[10, 6, 4]]], [[[2, 1, 7]]]])
    sigma = np.full(np.shape(x1), 1)
    distribution = ElectronSampler(np.array([[1, 1, 3], [3, 2, 3]]), f)
    move_probability = distribution.log_prob_gaussian(x1, x2, sigma)
    assert (move_probability == np.array([-49, -10.5])).all()


def test_steps():

    # test for gauss_initialize_position
    distribution = ElectronSampler(np.array([[1, 1, 3], [3, 2, 3]]),
                                   f,
                                   batch_no=2,
                                   steps=1000)
    distribution.gauss_initialize_position(np.array([[1], [2]]))
    assert ((distribution.x -
             np.array([[[[1, 1, 3]], [[3, 2, 3]], [[3, 2, 3]]],
                       [[[1, 1, 3]], [[3, 2, 3]], [[3, 2, 3]]]])) != 0).any()

    # testing symmetric simultaneous_move
    x1 = distribution.x
    distribution.move()
    assert ((distribution.x - x1) != 0).all()

    # testing asymmetric simultaneous_move
    distribution = ElectronSampler(np.array([[1, 1, 3], [3, 2, 3]]),
                                   f,
                                   batch_no=2,
                                   steps=1000,
                                   symmetric=False)
    distribution.gauss_initialize_position(np.array([[1], [2]]))
    x1 = distribution.x
    distribution.move(asymmetric_func=distribution.harmonic_mean)
    assert ((distribution.x - x1) != 0).all()
    assert np.shape(distribution.sampled_electrons) == (2000, 3, 1, 3)

    # testing symmetric single_move
    distribution = ElectronSampler(np.array([[1, 1, 3], [3, 2, 3]]),
                                   f,
                                   batch_no=2,
                                   steps=1000,
                                   simultaneous=False)
    distribution.gauss_initialize_position(np.array([[1], [2]]))
    x1 = distribution.x
    distribution.move(index=1)
    assert ((distribution.x[:, 1, :, :] - x1[:, 1, :, :]) != 0).all()
    assert ((distribution.x[:, 2, :, :] - x1[:, 2, :, :]) == 0).all()
    assert np.shape(distribution.sampled_electrons) == (2000, 3, 1, 3)

    # testing asymmetric single_move
    distribution = ElectronSampler(np.array([[1, 1, 3], [3, 2, 3]]),
                                   f,
                                   batch_no=2,
                                   steps=1000,
                                   simultaneous=False,
                                   symmetric=False)
    distribution.gauss_initialize_position(np.array([[1], [2]]))
    x1 = distribution.x
    distribution.move(asymmetric_func=distribution.harmonic_mean, index=1)
    assert ((distribution.x[:, 1, :, :] - x1[:, 1, :, :]) != 0).all()
    assert ((distribution.x[:, 2, :, :] - x1[:, 2, :, :]) == 0).all()
    assert np.shape(distribution.sampled_electrons) == (2000, 3, 1, 3)



import logging
from typing import List

import numpy as np

from deepchem.utils.typing import RDKitMol
from deepchem.utils.molecule_feature_utils import one_hot_encode
from deepchem.feat.base_classes import Featurizer
from typing import Any, Iterable, Optional

logger = logging.getLogger(__name__)

ZINC_CHARSET = [
    '#', ')', '(', '+', '-', '/', '1', '3', '2', '5', '4', '7', '6', '8', '=',
    '@', 'C', 'B', 'F', 'I', 'H', 'O', 'N', 'S', '[', ']', '\\', 'c', 'l', 'o',
    'n', 'p', 's', 'r'
]


class OneHotFeaturizer(Featurizer):
    """Encodes any arbitrary string or molecule as a one-hot array.

    This featurizer encodes the characters within any given string as a one-hot
    array. It also works with RDKit molecules: it can convert RDKit molecules to
    SMILES strings and then one-hot encode the characters in said strings.

    Standalone Usage:

    >>> import deepchem as dc
    >>> featurizer = dc.feat.OneHotFeaturizer()
    >>> smiles = ['CCC']
    >>> encodings = featurizer.featurize(smiles)
    >>> type(encodings[0])
    <class 'numpy.ndarray'>
    >>> encodings[0].shape
    (100, 35)
    >>> featurizer.untransform(encodings[0])
    'CCC'

    Note
    ----
    This class needs RDKit to be installed in order to accept RDKit molecules as
    inputs.

    It does not need RDKit to be installed to work with arbitrary strings.

    """

    def __init__(self,
                 charset: List[str] = ZINC_CHARSET,
                 max_length: Optional[int] = 100):
        """Initialize featurizer.

        Parameters
        ----------
        charset: List[str] (default ZINC_CHARSET)
            A list of strings, where each string is length 1 and unique.
        max_length: Optional[int], optional (default 100)
            The max length for string. If the length of string is shorter than
            max_length, the string is padded using space.

        If max_length is None, no padding is performed and arbitrary length
        strings are allowed.
        """
        if len(charset) != len(set(charset)):
            raise ValueError("All values in charset must be unique.")
        self.charset = charset
        self.max_length = Optional[int]
        if max_length is not None:
            self.max_length = int(max_length)
        else:
            self.max_length = None

    def featurize(self,
                  datapoints: Iterable[Any],
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """Featurize strings or mols.

        Parameters
        ----------
        datapoints: list
            A list of either strings (str or numpy.str_) or RDKit molecules.
        log_every_n: int, optional (default 1000)
            How many elements are featurized every time a featurization is logged.

        """
        datapoints = list(datapoints)
        if (len(datapoints) < 1):
            return np.array([])
        # Featurize data using featurize() in parent class
        return Featurizer.featurize(self, datapoints, log_every_n)

    def _featurize(self, datapoint: Any, **kwargs):
        # Featurize str data
        if isinstance(datapoint, (str, np.str_)):
            return self._featurize_string(datapoint)
        # Featurize mol data
        else:
            return self._featurize_mol(datapoint)

    def _featurize_string(self, string: str) -> np.ndarray:
        """Compute one-hot featurization of string.

        Parameters
        ----------
        string: str
            An arbitrary string to be featurized.

        Returns
        -------
        np.ndarray
            An one hot vector encoded from arbitrary input string.
            The shape is `(max_length, len(charset) + 1)`.
            The index of unknown character is `len(charset)`.

        """
        if isinstance(self.max_length, int):
            if (len(string) > self.max_length):  # Validation
                raise ValueError(
                    "The length of {} is longer than `max_length`.")
            string = self.pad_string(string)  # Padding

        return np.array([
            one_hot_encode(val, self.charset, include_unknown_set=True)
            for val in string
        ])

    def _featurize_mol(self, mol: RDKitMol) -> np.ndarray:
        """Compute one-hot featurization of this molecule.

        Parameters
        ----------
        mol: rdKit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            An one hot vector encoded from SMILES.
            The shape is '(max_length, len(charset) + 1)'
            The index of unknown character is 'len(charset)'.

        """
        try:
            from rdkit import Chem
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")
        smiles = Chem.MolToSmiles(mol)  # Convert mol to SMILES string.
        return self._featurize_string(smiles)  # Use string featurization.

    def pad_smile(self, smiles: str) -> str:
        """Pad SMILES string to `self.pad_length`

        Parameters
        ----------
        smiles: str
            The SMILES string to be padded.

        Returns
        -------
        str
            SMILES string space padded to self.pad_length

        """
        return self.pad_string(smiles)

    def pad_string(self, string: str) -> str:
        """Pad string to `self.pad_length`

        Parameters
        ----------
        string: str
            The string to be padded.

        Returns
        -------
        str
            String space padded to self.pad_length

        """
        if isinstance(self.max_length, int):
            return string.ljust(self.max_length)
        else:
            return string

    def untransform(self, one_hot_vectors: np.ndarray) -> str:
        """Convert from one hot representation back to original string

        Parameters
        ----------
        one_hot_vectors: np.ndarray
            An array of one hot encoded features.

        Returns
        -------
        str
            Original string for an one hot encoded array.

        """
        string = ""
        for one_hot in one_hot_vectors:
            try:
                idx = np.argmax(one_hot)
                string += self.charset[idx]
            except IndexError:
                string += ""
        return string



"""
Utility Functions for computing features on batch.
"""
import numpy as np


def batch_coulomb_matrix_features(X_b: np.ndarray,
                                  distance_max: float = -1,
                                  distance_min: float = 18,
                                  n_distance: int = 100):
    """Computes the values for different Feature on given batch.
    It works as a helper function to coulomb matrix.

    This function takes in a batch of Molecules represented as Coulomb Matrix.

    It proceeds as follows:

    - It calculates the Number of atoms per molecule by counting all the non zero elements(numbers) of every\
    molecule layer in matrix in one dimension.

    - The Gaussian distance is calculated using the Euclidean distance between the Cartesian coordinates of two atoms.\
    The distance value is then passed through a Gaussian function, which transforms it into a continuous value.

    - Then using number of atom per molecule, calculates the atomic charge by looping over the molecule layer in the Coulomb matrix\
    and takes the `2.4` root of the diagonal of `2X` of each molecule layer. `Undoing the Equation of coulomb matrix.`

    - Atom_membership is assigned as a commomn repeating integers for all the atoms for a specific molecule.

    - Distance Membership encodes spatial information, assigning closer values to atoms that are in that specific molecule.\
    All initial Distances are added a start value to them which are unique to each molecule.

    Models Used in:

    * DTNN

    Parameters
    ----------
    X_b: np.ndarray
        It is a 3d Matrix containing information of each the atom's ionic interaction with other atoms in the molecule.
    distance_min: float (default -1)
        minimum distance of atom pairs (in Angstrom)
    distance_max: float (default = 18)
        maximum distance of atom pairs (in Angstrom)
    n_distance: int (default 100)
        granularity of distance matrix
        step size will be (distance_max-distance_min)/n_distance

    Returns
    -------
    atom_number: np.ndarray
        Atom numbers are assigned to each atom based on their atomic properties.
        The atomic numbers are derived from the periodic table of elements.
        For example, hydrogen -> 1, carbon -> 6, and oxygen -> 8.
    gaussian_dist: np.ndarray
        Gaussian distance refers to the method of representing the pairwise distances between atoms in a molecule using Gaussian functions.
        The Gaussian distance is calculated using the Euclidean distance between the Cartesian coordinates of two atoms.
        The distance value is then passed through a Gaussian function, which transforms it into a continuous value.
    atom_mem: np.ndarray
        Atom membership refers to the binary representation of whether an atom belongs to a specific group or property within a molecule.
        It allows the model to incorporate domain-specific information and enhance its understanding of the molecule's properties and interactions.
    dist_mem_i: np.ndarray
        Distance membership i are utilized to encode spatial information and capture the influence of atom distances on the properties and interactions within a molecule.
        The inner membership function assigns higher values to atoms that are closer to the atoms' interaction region, thereby emphasizing the impact of nearby atoms.
    dist_mem_j: np.ndarray
        It captures the long-range effects and influences between atoms that are not in direct proximity but still contribute to the overall molecular properties.
        Distance membership j are utilized to encode spatial information and capture the influence of atom distances on the properties and interactions outside a molecule.
        The outer membership function assigns higher values to atoms that are farther to the atoms' interaction region, thereby emphasizing the impact of farther atoms.

    Examples
    --------
    >>> import os
    >>> import deepchem as dc
    >>> current_dir = os.path.dirname(os.path.abspath(__file__))
    >>> dataset_file = os.path.join(current_dir, 'test/assets/qm9_mini.sdf')
    >>> TASKS = ["alpha", "homo"]
    >>> loader = dc.data.SDFLoader(tasks=TASKS,
    ...                            featurizer=dc.feat.CoulombMatrix(29),
    ...                            sanitize=True)
    >>> data = loader.create_dataset(dataset_file, shard_size=100)
    >>> inputs = dc.utils.batch_utils.batch_coulomb_matrix_features(data.X)

    References
    ----------
    .. [1] Montavon, Grégoire, et al. "Learning invariant representations of
        molecules for atomization energy prediction." Advances in neural information
        processing systems. 2012.

    """
    distance = []
    atom_membership = []
    distance_membership_i = []
    distance_membership_j = []

    # Calculation of Step Size and steps
    step_size = (distance_max - distance_min) / n_distance
    steps = np.array([distance_min + i * step_size for i in range(n_distance)])
    steps = np.expand_dims(steps, 0)

    # Number of atoms per molecule is calculated by counting all the non zero elements(numbers) of every molecule.
    num_atoms = list(map(sum, X_b.astype(bool)[:, :, 0]))

    # It loops over the molecules in the Coulomb matrix and takes the "2.4" root of the diagonal of "2X" of each molecule's representation.
    atom_number = [
        np.round(
            np.power(2 * np.diag(X_b[i, :num_atoms[i], :num_atoms[i]]),
                     1 / 2.4)).astype(int) for i in range(len(num_atoms))
    ]
    start = 0
    for im, molecule in enumerate(atom_number):
        distance_matrix = np.outer(
            molecule, molecule) / X_b[im, :num_atoms[im], :num_atoms[im]]
        np.fill_diagonal(distance_matrix, -100)
        distance.append(np.expand_dims(distance_matrix.flatten(), 1))
        atom_membership.append([im] * num_atoms[im])
        membership = np.array([np.arange(num_atoms[im])] * num_atoms[im])
        membership_i = membership.flatten(order='F')
        membership_j = membership.flatten()
        distance_membership_i.append(membership_i + start)
        distance_membership_j.append(membership_j + start)
        start = start + num_atoms[im]
    atom_number = np.concatenate(atom_number).astype(np.int32)
    distance = np.concatenate(distance, axis=0)

    # Calculates the Gaussian Distance by passing distance by a gaussian function.
    gaussian_dist = np.exp(-np.square(distance - steps) / (2 * step_size**2))
    gaussian_dist = gaussian_dist.astype(np.float64)
    atom_mem = np.concatenate(atom_membership).astype(np.int64)
    dist_mem_i = np.concatenate(distance_membership_i).astype(np.int64)
    dist_mem_j = np.concatenate(distance_membership_j).astype(np.int64)
    features = [atom_number, gaussian_dist, atom_mem, dist_mem_i, dist_mem_j]
    return features



"""Normalizing flows for transforming probability distributions using PyTorch.
"""
import torch
from torch import nn
from typing import Sequence, Tuple


class NormalizingFlow(nn.Module):
    """Normalizing flows are widley used to perform generative models.
    This algorithm gives advantages over variational autoencoders (VAE) because
    of ease in sampling by applying invertible transformations
    (Frey, Gadepally, & Ramsundar, 2022).

    Example
    --------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.layers import Affine
    >>> from deepchem.models.torch_models.normalizing_flows_pytorch import NormalizingFlow
    >>> import torch
    >>> from torch.distributions import MultivariateNormal
    >>> # initialize the transformation layer's parameters
    >>> dim = 2
    >>> samples = 96
    >>> transforms = [Affine(dim)]
    >>> distribution = MultivariateNormal(torch.zeros(dim), torch.eye(dim))
    >>> # initialize normalizing flow model
    >>> model = NormalizingFlow(transforms, distribution, dim)
    >>> # evaluate the log_prob when applying the transformation layers
    >>> input = distribution.sample(torch.Size((samples, dim)))
    >>> len(model.log_prob(input))
    96
    >>> # evaluates the the sampling method and its log_prob
    >>> len(model.sample(samples))
    2

    """

    def __init__(self, transform: Sequence, base_distribution,
                 dim: int) -> None:
        """This class considers a transformation, or a composition of transformations
        functions (layers), between a base distribution and a target distribution.

        Parameters
        ----------
        transform: Sequence
            Bijective transformation/transformations which are considered the layers
            of a Normalizing Flow model.
        base_distribution: torch.Tensor
            Probability distribution to initialize the algorithm. The Multivariate Normal
            distribution is mainly used for this parameter.
        dim: int
            Value of the Nth dimension of the dataset.

        """
        super().__init__()
        self.dim = dim
        self.transforms = nn.ModuleList(transform)
        self.base_distribution = base_distribution

    def log_prob(self, inputs: torch.Tensor) -> torch.Tensor:
        """This method computes the probability of the inputs when
        transformation/transformations are applied.

        Parameters
        ----------
        inputs: torch.Tensor
            Tensor used to evaluate the log_prob computation of the learned
            distribution.
            shape: (samples, dim)

        Returns
        -------
        log_prob: torch.Tensor
            This tensor contains the value of the log probability computed.
            shape: (samples)

        """
        log_prob = torch.zeros(inputs.shape[0])
        for biject in reversed(self.transforms):
            inputs, inverse_log_det_jacobian = biject.inverse(inputs)
            log_prob += inverse_log_det_jacobian

        return log_prob

    def sample(self, n_samples: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """Performs a sampling from the transformed distribution.
        Besides the outputs (sampling), this method returns the logarithm of
        probability to obtain the outputs at the base distribution.

        Parameters
        ----------
        n_samples: int
            Number of samples to select from the transformed distribution

        Returns
        -------
        sample: tuple
            This tuple contains a two torch.Tensor objects. The first represents
            a sampling of the learned distribution when transformations had been
            applied. The secong torc.Tensor is the computation of log probabilities
            of the transformed distribution.
            shape: ((samples, dim), (samples))

        """
        outputs = self.base_distribution.sample((n_samples,))
        log_prob = self.base_distribution.log_prob(outputs)

        for biject in self.transforms:
            outputs, log_det_jacobian = biject.forward(outputs)
            log_prob += log_det_jacobian

        return outputs, log_prob
